/*******************************************************************
 *
 *	File:		GameEngine.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include "EikEnv.h"
#include <stdio.h>
#include <unistd.h>

#include "GameEngine.h"

const TInt KGameStackSize	= 0x10000;			// 64KB
const TInt KGameHeapSizeMin	= 0x10000;			// 64KB
const TInt KGameHeapSizeMax	= 0x2000000;		// 32 MB

#define GAME_DLL	_L("edoom.dll")


CGameRunner::~CGameRunner()
{
	Cancel();

#ifdef __WINS__
	if (iGameLib.Handle())
		iGameLib.Close();
#endif
}

CGameRunner::CGameRunner(MGameWatcher& aGameWatcher, CGameEngine& aGameEngine) : CActive(CActive::EPriorityStandard), iGameWatcher(aGameWatcher), iGameEngine(aGameEngine)
{
}

CGameRunner* CGameRunner::NewL(MGameWatcher& aGameWatcher, CGameEngine& aGameEngine)
{
	CGameRunner*	self = new(ELeave) CGameRunner(aGameWatcher, aGameEngine);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();			// self
	return self;
}

void CGameRunner::ConstructL()
{
	CActiveScheduler::Add(this);
#ifdef __WINS__
	User::LeaveIfError(iGameLib.Load(GAME_DLL));
	iMainFunc = (TMainFunc) iGameLib.Lookup(1);
#endif
}

void CGameRunner::RunL()
{
	iGameWatcher.NotifyGameFinished();
}

void CGameRunner::DoCancel()
{
	iChild.LogonCancel(iStatus);
}


#define EXE_ON_C "C:\\egnuboy\\GNUBOY.EXE"
#define EXE_ON_D "D:\\egnuboy\\GNUBOY.EXE"

void CGameRunner::RunGameL(const TDesC& aGameName, const TGameOptions& aGameOptions)
{
	iGameName = aGameName;
	iFileName.Copy(iGameName);
	
	iArgc = 0;
	iArgv[iArgc++] = (access(EXE_ON_D, 0) != -1) ? EXE_ON_D : EXE_ON_C;
	iArgv[iArgc++] = "--savedir=c:\\egnuboy";
	iArgv[iArgc++] = (char*) iFileName.PtrZ();
	if (aGameOptions.iGameSize == TGameOptions::EGameSizeFitScreen)
		iArgv[iArgc++] = "--fitscreen=1";
	iArgv[iArgc] = NULL;

	iCommandLine.Copy(_L(""));
	for(TInt i=1 ; i<iArgc ; i++)
	{
		iCommandLine.Append(TPtrC8((const TUint8*)iArgv[i]));
		iCommandLine.Append(_L(" "));
	}
	RDebug::Print(_L("Cmdline: %S"), &iCommandLine);

	StartGameL(aGameName);

	iChild.Logon(iStatus);
	SetActive();

	iChild.Resume();
}



#ifdef __WINS__
void CGameRunner::StartGameL(const TDesC& aGameName)
{
	TBuf<32>	threadName;

		/*
		 * Hmm, there is something funny about the game thread on target. Eventhough the
		 * previous game thread has finished (we even called Terminate on it) we cannot
		 * create a new instance of it with the same name (returns with KErrAlreadyExists).
		 * Using unique names gets round that.
		 */
	threadName.Format(_L("Game%d"), iGameCounter++);
	User::LeaveIfError(iChild.Create(threadName, CGameRunner::ThreadGame,  KGameStackSize, KGameHeapSizeMin, KGameHeapSizeMax, this, EOwnerThread));

}

TInt CGameRunner::ThreadGame(TAny* aData)
{  
	CTrapCleanup* cleanup=CTrapCleanup::New();

	CGameRunner*	self = (CGameRunner*) aData;
	TRAPD(err, self->DoRunGameL());

	delete cleanup;

	return(err);
}

void CGameRunner::DoRunGameL()
{
	(*iMainFunc)(iArgc, iArgv);

}

#else

void CGameRunner::StartGameL(const TDesC& aGameName)
{
	TPtrC	fileName((const TUint8*)iArgv[0]);
	User::LeaveIfError(iChild.Create(fileName, iCommandLine));
}

#endif





void CGameRunner::TerminateGame()
{
	Cancel();
	ShutdownGame();
}


void CGameRunner::ShutdownGame()
{
	if (iChild.Handle())
	{
#ifdef __WINS__
			/*
			 * Throw in some waits before killing it off. Thread's dying whilst running under the  
			 * may result in a emulator + devstudio freeze.
			 */
		User::After(1000 * 500);
		iGameEngine.KillRunningStdoutConsole();
		User::After(1000 * 500);
#else
		iChild.Terminate(0);
#endif
		iChild.Close();
	}
}


CGameEngine* CGameEngine::NewL(RFs& aFs)
{
	CGameEngine* self = new CGameEngine(aFs);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}

CGameEngine::CGameEngine(RFs& aFs) : iFs(aFs)
{
}

CGameEngine::~CGameEngine()
{
	if (iGlobalDataChunk.Handle())
		iGlobalDataChunk.Close();
}


void CGameEngine::ConstructL()
{
	User::LeaveIfError(iGlobalDataChunk.CreateGlobal(KGlobalDataChunkName, 4, 4));

	iGameNameArray = new(ELeave)CArrayFixFlat<TGameName>(5);

	CDir* entryList = NULL;
	iFs.GetDir(_L("c:\\egnuboy\\roms\\*.gb"), 0, 0, entryList);
	CleanupStack::PushL(entryList);
	if (entryList)
		AddGames(entryList);
	CleanupStack::PopAndDestroy();		// entryList

	entryList = NULL;
	iFs.GetDir(_L("c:\\egnuboy\\roms\\*.gbc"), 0, 0, entryList);
	CleanupStack::PushL(entryList);
	if (entryList)
		AddGames(entryList);
	CleanupStack::PopAndDestroy();		// entryList

	entryList = NULL;
	iFs.GetDir(_L("d:\\egnuboy\\roms\\*.gb"), 0, 0, entryList);
	CleanupStack::PushL(entryList);
	if (entryList)
		AddGames(entryList);
	CleanupStack::PopAndDestroy();		// entryList

	entryList = NULL;
	iFs.GetDir(_L("d:\\egnuboy\\roms\\*.gbc"), 0, 0, entryList);
	CleanupStack::PushL(entryList);
	if (entryList)
		AddGames(entryList);
	CleanupStack::PopAndDestroy();		// entryList
}





void CGameEngine::KillRunningStdoutConsole()
{
	TInt	id = 0;
	TBool	sentEvent = EFalse;
	RWsSession&		wsSession = CEikonEnv::Static()->WsSession();

		/*
		 * When EPOC's STDLIB shuts down it will put up a "press a key to close console".
		 * In WINS, we'll find the window group belonging to the thread STDOUT and send
		 * it a key event to have it automatically closed.
		 */
	while (id >= 0)
	{
		id = wsSession.FindWindowGroupIdentifier(id , _L("*"));
		if (id > 0)
		{
			TThreadId threadId;
			if (wsSession.GetWindowGroupClientThreadId(id , threadId) == KErrNone)
			{
				RThread	thread;
				if (thread.Open(threadId) == KErrNone)
				{
					TFullName	name;
					name = thread.Name();
					if (name.Compare(_L("STDOUT")) == 0)
					{
						TWsEvent	event;
						TKeyEvent	key;
						
						key.iCode=EKeyEscape;
						key.iScanCode=0;
						key.iModifiers=0;

						event.SetType(EEventKey);
						*event.Key()=key;
						event.SetTimeNow();
						wsSession.SendEventToWindowGroup(id, event);

						sentEvent = ETrue;
					}
					thread.Close();
					RDebug::Print(_L("CGameAppUi: thread=%S, id=%d"), &name, id);
				}
			}
		}
	}
	if (!sentEvent)
		User::InfoPrint(_L("Please close the task named STDOUT"));
}




void CGameEngine::SetWindowGroupId(TInt aWindowGroupId)
{
}



TInt CGameEngine::GameNo(const TDesC& /*aGameName*/)
{
	TInt	gameNo = 0;
	return gameNo;
}

TBool CGameEngine::IsGameCloneOf(struct GameDriver* /*aGame*/)
{
	return EFalse;
}

TInt CGameEngine::GameCount()
{
	return iGameNameArray->Count();
}

TDesC& CGameEngine::GameName(TInt aGameNo)
{
	return (*iGameNameArray)[aGameNo];
}

void CGameEngine::ForceGameExit()
{
}


void CGameEngine::AddGames(CDir* aEntryList)
{
	for(TInt i=0; i<aEntryList->Count(); i++)
	{
		const TEntry& entry = (*aEntryList)[i];
		TParsePtrC	parse(entry.iName);
		iGameNameArray->AppendL(parse.Name());
		RDebug::Print(_L("Game:%S"), &entry.iName);
	}
}

TBool CGameEngine::GetGameFileLocation(const TDesC& aGameName, TDes& aGameFileName)
{
	TUint	dummy;
	TInt	err;

	aGameFileName.Format(_L("c:\\egnuboy\\roms\\%S.gb"), &aGameName);
	if ((err = iFs.Att(aGameFileName, dummy)) == KErrNone)
		return ETrue;
	aGameFileName.Format(_L("c:\\egnuboy\\roms\\%S.gbc"), &aGameName);
	if ((err = iFs.Att(aGameFileName, dummy)) == KErrNone)
		return ETrue;
	aGameFileName.Format(_L("d:\\egnuboy\\roms\\%S.gb"), &aGameName);
	if ((err = iFs.Att(aGameFileName, dummy)) == KErrNone)
		return ETrue;
	aGameFileName.Format(_L("d:\\egnuboy\\roms\\%S.gbc"), &aGameName);
	if ((err = iFs.Att(aGameFileName, dummy)) == KErrNone)
		return ETrue;
	return EFalse;
}
