/*******************************************************************
 *
 *	File:		Video.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <stdio.h>
#include <math.h>
#include <gdi.h>
#include "video.h"
#include "game.h"


const KOneSecondInMicroSeconds =	1000 * 1000;

static TInt sThrottle				= 1;	// bool option, default: true
static TInt sManualDiffFrames		= 0;


/***********************************************
 *
 * BITMAP
 *
 ***********************************************/

CGameBitmap::CGameBitmap()
{
}

CGameBitmap* CGameBitmap::NewL(TInt aWidth, TInt aHeight, TDisplayMode aDisplayMode)
{
	CGameBitmap*	self = new(ELeave) CGameBitmap;
	CleanupStack::PushL(self);
	self->ConstructL(aWidth, aHeight, aDisplayMode);
	CleanupStack::Pop();
	return self;
}

CGameBitmap::~CGameBitmap()
{
	delete iBitmapContext;
	iBitmapContext = NULL;
	delete iBitmapDevice;
	iBitmapDevice = NULL;
	delete iBitmap;
	iBitmap = NULL;
	if (iOsdBitmap.line)
	{
		User::Free(iOsdBitmap.line);
		iOsdBitmap.line = NULL;
	}
}


TDisplayMode CGameBitmap::GetDisplayMode(TInt aBpp, TBool aColor)
{
	if (aBpp == 1)
		return EGray2;
	else if (aBpp == 2)
		return EGray4;
	else if (aBpp == 4)
		return aColor ? EColor16 : EGray16;
	else if (aBpp == 8)
		return aColor ? EColor256 : EGray256;
	else if (aBpp == 16)
		return EColor64K;
	else if (aBpp == 12)
		return EColor4K;
	else if (aBpp == 24)
		return EColor16M;
	else if (aBpp == 32)
		return ERgb;

	ASSERT(EFalse);
	return ENone;
}


/*
 * GetBpp returns the number of bits we need per pixel in the given
 * display mode. It is different from GetColorDepth in that GetBpp can be
 * used to calculate byte offsets in a bitmap/framebuffer.
 * For example, GetBpp(EColor4K) returns 16 because each pixel requires
 * 16 bits (of which only the least significant 12 bits are used).
 * GetColorDepth(EColor4K) returns 12.
 */
TInt CGameBitmap::GetBpp(TDisplayMode aDisplayMode)
{
	const TInt bpp[] =
		{
			-1,
			1,
			2,
			4,
			8,
			4,
			8,
			16,
			32,
			32,
			16
		};
	return bpp[aDisplayMode];
}

TInt CGameBitmap::GetColorDepth(TDisplayMode aDisplayMode)
{
	const TInt bpp[] =
		{
			-1,
			1,
			2,
			4,
			8,
			4,
			8,
			16,
			24,
			32,
			12
		};
	return bpp[aDisplayMode];
}


void CGameBitmap::ConstructL(TInt aWidth, TInt aHeight, TDisplayMode aDisplayMode)
{
	iBitmap = new (ELeave) CGameFbsBitmap;
	iBitmap->Create(TSize(aWidth, aHeight), aDisplayMode);

	iBitmapDevice = CFbsBitmapDevice::NewL(iBitmap);
	iBitmapDevice->CreateContext(iBitmapContext);

		/*
		 * Setup OsdBitmap as used by the 
		 */
	iOsdBitmap.width	= aWidth;
	iOsdBitmap.height	= aHeight;
	iOsdBitmap.depth	= 8; // GetColorDepth(aDisplayMode);
	iOsdBitmap._private = (TAny *) this;
	iOsdBitmap.line		= (TUint8 **) User::Alloc(aHeight * sizeof(TUint8 *));
	User::LeaveIfNull(iOsdBitmap.line);

	TUint32*			dataAddress = iBitmap->DataAddress();
	CBitwiseBitmap*		bitwiseBitmap =iBitmap->GameAddress();
	for (TInt i=0 ; i<aHeight ; i++)
		iOsdBitmap.line[i] = (TUint8 *) bitwiseBitmap->ScanLineAddress(dataAddress, i);

		/*
		 * Setup iOsdVisibleBitmap, it's mapped as a rect into iOsdBitmap
		 */
	iOsdVisibleBitmap.width		= 0;
	iOsdVisibleBitmap.height	= 0;
	iOsdVisibleBitmap.depth		= 8; // GetColorDepth(aDisplayMode);
	iOsdVisibleBitmap._private	= (TAny *) this;
	iOsdVisibleBitmap.line		= NULL;

	SetVisibleAreaL(TRect(0, 0, aWidth, aHeight));
}

void CGameBitmap::SetVisibleAreaL(const TRect& aVisibleArea)
{
	TInt	width = aVisibleArea.Size().iWidth;
	TInt	height = aVisibleArea.Size().iHeight;

	iOsdVisibleBitmap.width		= width;
	iOsdVisibleBitmap.height	= height;

	if (iOsdVisibleBitmap.line)
	{
		User::Free(iOsdVisibleBitmap.line);
		iOsdVisibleBitmap.line = NULL;
	}
	iOsdVisibleBitmap.line = (TUint8 **) User::Alloc(height * sizeof(TUint8 *));;
	for (TInt h=0 ; h<height ; h++)
		iOsdVisibleBitmap.line[h] = iOsdBitmap.line[h + aVisibleArea.iTl.iY] + ((aVisibleArea.iTl.iX * iOsdVisibleBitmap.depth) / 8);
}




/***********************************************
 *
 * SCREEN
 *
 ***********************************************/

CGameScreen::CGameScreen(const TScreenInfoV01& aScreenInfo, TDisplayMode aDisplayMode)
{
	iHasFrameBuffer		= aScreenInfo.iScreenAddressValid;
	iFrameBuffer		= iHasFrameBuffer ? (TUint8*) aScreenInfo.iScreenAddress : NULL;
	iBytesPerScanLine	= (aScreenInfo.iScreenSize.iWidth * CGameBitmap::GetBpp(aDisplayMode)) / 8;
	iScreenSize			= aScreenInfo.iScreenSize; 
	iDisplayMode		= aDisplayMode;

#ifndef __WINS__
#ifndef __ER6__
		/*
		 * Hmm, I've just checked out the SA1100 spec. It seems that for 8bpp displays
		 * there is a 512 palette table at the beginning of the frame buffer.
		 * This code assumes that all ER5 target build with 8bpp are SA1100 machines. 
		 * I only know 2: Series 7 and Netbook.
		 */
	if (CGameBitmap::GetBpp(aDisplayMode) == 8)
		iFrameBuffer += 512;
#endif
#endif

}

CGameScreen::~CGameScreen()
{
	User::Free(iScanLine);
}


void CGameScreen::ConstructL(TInt aWidth, TInt aHeight, TInt aDepth, TOrientation aOrientation, TInt aFps)
{
		/*
		 * game throttling
		 */
	TTimeIntervalMicroSeconds32	period;
	UserHal::TickPeriod(period);
	iPeriod				= period.Int();
	iVideoFps			= aFps;
	iDoubleFrame		= -1;
	iRealFrameCount		= 0;
	iGameFrameCount		= 0;
	iTicksGameStart		= 0;

	iWidth			= iGameWidth	= aWidth;
	iHeight			= iGameHeight	= aHeight;
	iGameDepth		= aDepth;
	iOrientation	= aOrientation;

	iScanLine = (TUint8*) User::AllocL(iBytesPerScanLine);
	Mem::FillZ(iScanLine, iBytesPerScanLine);

	SetupScaling();
	SetupPosition();

	SetupScreenAccess();

	iBytesPerBitmapLine = (iWidth * iGameDepth) / 8;
}

void CGameScreen::SetWantScaling(TBool aWantScaling)
{
	iWantScaling = aWantScaling;
}

void CGameScreen::SetWantDoubleBitmap(TBool aWantDoubleBitmap)
{
	iWantDoubleBitmap = aWantDoubleBitmap;
}


void CGameScreen::SetBitmapPosition(const TPoint& aPosition)
{
	TInt	bpp = CGameBitmap::GetBpp(iDisplayMode);
	iBitmapPosition = aPosition;
	iBitmapStartAddress = iFrameBuffer + ((aPosition.iX * bpp) / 8);
	iBitmapStartAddress += (aPosition.iY * iBytesPerScanLine);
}



void CGameScreen::SetupScreenAccess()
{
	if (iHasFrameBuffer)
	{
		if ((iPixelSkipX == -1) && (iPixelSkipY == -1))
		{
			if (iDoubleBitmap)
			{
				const TBitmapDrawer	bitmapDrawers[] =
					{
						NULL,
						NULL,											// Gray2
						NULL,											// Gray4
						NULL,											// Gray16
						&CGameScreen::DirectDrawDoubleBitmap8bppTo8bpp,	// Gray256
						NULL,											// Color16
						&CGameScreen::DirectDrawDoubleBitmap8bppTo8bpp,	// Color256
						NULL,											// Color64K
						NULL,											// Color16M
						NULL,											// Rgb
						NULL											// Color4K
					};
				iBitmapDrawer = bitmapDrawers[iDisplayMode];
				iHasFrameBuffer = (iBitmapDrawer != NULL);
			}
			else
			{
				const TBitmapDrawer	bitmapDrawers[] =
					{
						NULL,
						NULL,										// Gray2
						&CGameScreen::DirectDrawBitmap8bppTo2bpp,	// Gray4
						&CGameScreen::DirectDrawBitmap8bppTo4bpp,	// Gray16
						&CGameScreen::DirectDrawBitmap8bppTo8bpp,	// Gray256
						&CGameScreen::DirectDrawBitmap8bppTo4bpp,	// Color16
						&CGameScreen::DirectDrawBitmap8bppTo8bpp,	// Color256
						NULL,										// Color64K
						NULL,										// Color16M
						NULL,										// Rgb
						&CGameScreen::DirectDrawBitmap8bppTo12bpp	// Color4K
					};
				iBitmapDrawer = bitmapDrawers[iDisplayMode];
				iHasFrameBuffer = (iBitmapDrawer != NULL);
			}
		}
		else
		{
			if (iDoubleBitmap)
			{
				const TScaledScanLineDrawer	scaledScanLineDrawers[] =
					{
						NULL,
						NULL,												// Gray2
						NULL,												// Gray4
						NULL,												// Gray16
						&CGameScreen::DrawDoubleScaledScanLine8bppTo8bpp,	// Gray256
						NULL,												// Color16
						&CGameScreen::DrawDoubleScaledScanLine8bppTo8bpp,	// Color256
						NULL,												// Color64K
						NULL,												// Color16M
						NULL,												// Rgb
						NULL,												// Color4K
					};
				iBitmapDrawer = &CGameScreen::DoubleScaledDrawBitmap;
				iScaledScanLineDrawer = scaledScanLineDrawers[iDisplayMode];
				iHasFrameBuffer = (iScaledScanLineDrawer != NULL);
			}
			else
			{
				const TScaledScanLineDrawer	scaledScanLineDrawers[] =
					{
						NULL,
						NULL,												// Gray2
						DrawScaledScanLine8bppTo2bpp,						// Gray4
						DrawScaledScanLine8bppTo4bpp,						// Gray16
						NULL,												// Gray256
						NULL,												// Color16
						NULL,												// Color256
						NULL,												// Color64K
						NULL,												// Color16M
						NULL,												// Rgb
						DrawScaledScanLine8bppTo12bpp,						// Color4K
					};
				iBitmapDrawer = &CGameScreen::ScaledDrawBitmap;
				iScaledScanLineDrawer = scaledScanLineDrawers[iDisplayMode];
				iHasFrameBuffer = (iScaledScanLineDrawer != NULL);
			}
		}
	}
}

#define KMaxDoubleFrame 5

static sDoubleFrame[KMaxDoubleFrame];
static sDoubleFrameCount[KMaxDoubleFrame];


TInt CGameScreen::SkipThisFrame()
{
	return 0;
}

/*
 * Implementing throttling isn't that easy in EPOC because EPOC timers aren't that 
 * fine grained. The smallest period you can sleep is roughly 40ms and that isn't
 * good enough. Let's call the frame rate at which the hardware is capable of running 
 * "real frame rate" and the rate at which the game should run "game frame rate". 
 * 
 * An example. Suppose the game frame rate is 50 fps and the real frame rate is 60 fps. 
 * 50fps means that each frame should take 1000/50 = 20ms and at 60 fps, each frame 
 * only takes 1000/60 = 16.67 ms. So, in this case we should sleep for 2.33 ms inbetween 
 * drawing each frame to get the correct 50fps.
 *
 * Since we cannot sleep, what we'll do instead is take advantage of a known factor:
 * the time it takes to draw a frame. In the above example, if we run at 60 fps then
 * if we draw 10 frames twice in each second, the games runs at the correct speed
 * of 50 fps as well.
 */
#if 0
TInt CGameScreen::DoubleFrame(TInt aShowFps, struct osd_bitmap *bitmap)
{
	TInt	doubleFrame = 1;

		/*
		 * Initialise on first tick
		 */
	if (iTicksGameStart == 0)
	{
		iTotalGameFrames = iTotalRealFrames = 0;
		iTicksGameStart = User::TickCount();
		iPrevTicks = iTicksGameStart;
		iDoubleFrame = -1;
		Mem::Fill(sDoubleFrame, KMaxDoubleFrame * sizeof(TInt), 0);		// set to -1
		sDoubleFrameCount[0] = 1;
		sDoubleFrameCount[1] = 1;
		sDoubleFrameCount[2] = 1;
		sDoubleFrameCount[3] = 1;
		sDoubleFrameCount[4] = 1;

		iDiffFrames = 0;
		iRealFps	= 0;
		iGameFps	= 0;

		sManualDiffFrames		= 0;
	}
	if (sThrottle)
	{
		if (iGameFrameCount == iVideoFps)
		{
			TInt ticks = User::TickCount();
			TInt diff = ticks - iPrevTicks;
			iPrevTicks = ticks;

			TInt time = diff * iPeriod;
			if (!time)
				time = 1;
			iRealFps = (iRealFrameCount * KOneSecondInMicroSeconds) / time;
			iGameFps = (iGameFrameCount * KOneSecondInMicroSeconds) / time;

			if (time < KOneSecondInMicroSeconds)
				iDiffFrames = iRealFps - iVideoFps;
			else
				iDiffFrames -= (iVideoFps - iGameFps) / 2;

			TInt diffFrames = iDiffFrames;

			if (sManualDiffFrames > 0)
				diffFrames = sManualDiffFrames;

			Mem::Fill(sDoubleFrame, KMaxDoubleFrame * sizeof(TInt), 0xff);
			sDoubleFrameCount[0] = 1;
			sDoubleFrameCount[1] = 1;
			sDoubleFrameCount[2] = 1;
			sDoubleFrameCount[3] = 1;
			sDoubleFrameCount[4] = 1;
			for (TInt i=0 ; i<KMaxDoubleFrame && diffFrames; i++)
			{
				sDoubleFrame[i] = iVideoFps / diffFrames;
				if (sDoubleFrame[i] == 0) sDoubleFrame[i]=1;
				if (iVideoFps / sDoubleFrame[i] > diffFrames + 2)
					sDoubleFrame[i]++;
				if (sDoubleFrame[i] == 0)
					sDoubleFrame[i] = 1;
				diffFrames -= iVideoFps / sDoubleFrame[i];
			}
//			RDebug::Print(HERE("diff=%d, %d %d %d %d %d"), iDiffFrames, sDoubleFrame[0], sDoubleFrame[1], sDoubleFrame[2], sDoubleFrame[3], sDoubleFrame[4]);

			iGameFrameCount = 0;
			iRealFrameCount = 0;
		}
	

		for (TInt i=0 ; i<KMaxDoubleFrame ; i++)
		{
			if (sDoubleFrameCount[i] == sDoubleFrame[i])
			{
				doubleFrame++;
				sDoubleFrameCount[i] = 1;
				iRealFrameCount++;
				iTotalRealFrames++;
			}
			else
				sDoubleFrameCount[i]++;
		}

		iGameFrameCount++;
		iRealFrameCount++;
	}
	iTotalRealFrames++;
	iTotalGameFrames++;

	if (aShowFps)
	{
		static	sTotalRealFrames=0;
		static	sTotalGameFrames=0;
		static	sTotalGameFps=0;
		static	sTotalRealFps=0;
		static	sTotalTicks=0;

		char	fps[64];
		TInt	ticks = User::TickCount() - iTicksGameStart;
		TInt	seconds = (ticks * iPeriod) / (KOneSecondInMicroSeconds);

		if (seconds == 0)
			seconds = 1;

		if (seconds > 5)
		{
			sTotalRealFrames+=iTotalRealFrames;
			sTotalGameFrames+=iTotalGameFrames;
			sTotalTicks+=ticks;
			TInt	totSeconds = (sTotalTicks * iPeriod) / (KOneSecondInMicroSeconds);
			sTotalGameFps = sTotalGameFrames / totSeconds;
			sTotalRealFps = sTotalRealFrames / totSeconds;

			iTotalRealFrames=0;
			iTotalGameFrames=0;
			iTicksGameStart=User::TickCount();

			if (sManualDiffFrames > 0)
				sprintf(fps, "manual: g: %d, r: %d, (%d, %d) %d\n", iGameFps, iRealFps, sTotalGameFps, sTotalRealFps, sManualDiffFrames);
			else
				sprintf(fps, "auto: g: %d, r: %d, (%d, %d)\n", iGameFps, iRealFps, sTotalGameFps, sTotalRealFps);

			printf(fps);
			fflush(stdout);
		}

	}

	return doubleFrame;
}
#endif


void CGameScreen::FrameTick()
{
		/*
		 * Initialise on first tick
		 */
	if (iTicksGameStart == 0)
	{
		iTicksGameStart = User::TickCount();
		iPrevTicks = iTicksGameStart;
		iGameFps	= 0;
		iTotalGameFrames = iGameFrames	= 0;
	}
	iGameFrames++;

	TInt	curticks = User::TickCount();
	TInt	ticks = curticks - iPrevTicks;
	if ((ticks *iPeriod) > KOneSecondInMicroSeconds)
	{
		iGameFps = iGameFrames;
		iTotalGameFrames += iGameFrames;
		iGameFrames = 0;
		iPrevTicks = curticks;
	}
}



void CGameScreen::SetupPosition()
{
	iPosition = TPoint(0, 0);

	switch (iOrientation)
	{
		case EOrientationRotLeft:
			printf("SetupPosition: EOrientationRotLeft\n");
			iPosition = TPoint(0, 0);
			if ((iHeight +10) < iScreenSize.iHeight)
			{
				iPosition.iY = (iScreenSize.iHeight - iHeight) / 2;
			}
			break;
		case EOrientationRotRight:
			iPosition = TPoint(iScreenSize.iWidth - iWidth, 0);
			if ((iHeight +10) < iScreenSize.iHeight)
			{
				iPosition.iY = (iScreenSize.iHeight - iHeight) / 2;
			}
			break;
	
		default:
			if ((iWidth +10) < iScreenSize.iWidth)
			{
				iPosition.iX = (iScreenSize.iWidth - iWidth) / 2;
			}

			if ((iHeight +10) < iScreenSize.iHeight)
			{
				iPosition.iY = (iScreenSize.iHeight - iHeight) / 2;
			}
			break;
	}
	SetBitmapPosition(iPosition);
}

void CGameScreen::SetupScaling()
{
	iPixelSkipX = -1;
	iPixelSkipY = -1;
	iDoubleBitmap = EFalse;

	if (iWantScaling)
	{
		float	scaleX = ((float) iScreenSize.iWidth) / ((float) iGameWidth);
		float	scaleY = ((float) iScreenSize.iHeight) / ((float) iGameHeight);
		if ((scaleX < 1) || (scaleY < 1))
		{
			float scale = (scaleX < scaleY) ? scaleX : scaleY;
			iWidth = scale * iGameWidth;
			iHeight = scale * iGameHeight;

			iPixelSkipX = ceil (((float) iGameWidth) / (((float) iGameWidth) - ((float)iWidth)));
			iPixelSkipY = ceil (((float) iGameHeight) / (((float) iGameHeight) - ((float) iHeight)));

			iPixelSkipY--;	// we start counting at zero
			iPixelSkipX--;
		}
		else
		{
			TInt doubleWidth = iGameWidth * 2;
			TInt doubleHeight = iGameHeight * 2;

			float	scaleX = ((float) iScreenSize.iWidth) / ((float) doubleWidth);
			float	scaleY = ((float) iScreenSize.iHeight) / ((float) doubleHeight);
			if (((scaleX < 1) || (scaleY < 1)) && iWantDoubleBitmap)
			{
				iDoubleBitmap = ETrue;
				if ((scaleX > 0.66) && (scaleY > 0.66))
				{
					float scale = (scaleX < scaleY) ? scaleX : scaleY;
					iWidth = scale * doubleWidth;
					iHeight = scale * doubleHeight;

					iPixelSkipX = ceil (((float) doubleWidth) / (((float) doubleWidth) - ((float)iWidth)));
					iPixelSkipY = ceil (((float) doubleHeight) / (((float) doubleHeight) - ((float) iHeight)));

					iPixelSkipY--;
					iPixelSkipX--;
				}
				else
				{
					iDoubleBitmap = EFalse;
				}
			}
		}
	}
	else
	{
		iWantPanning = ETrue;		
	}
	iGameWidth-=4;
	iGameHeight-=4;
}


void CGameScreen::AllocateColors(TUint totalcolors, const TUint8 *palette, TUint16 *pens, TInt modifiable, const TUint8 *debug_palette,TUint16 *debug_pens)
{
	if (CGameBitmap::GetColorDepth(iDisplayMode) <= 8)
		AllocateEpocColors(totalcolors, palette, pens, modifiable, debug_palette, debug_pens);
	else
		AllocateGameColors(totalcolors, palette, pens, modifiable, debug_palette, debug_pens);
}

void CGameScreen::AllocateEpocColors(TUint totalcolors, const TUint8 *palette, TUint16 *pens, TInt modifiable, const TUint8 *debug_palette,TUint16 *debug_pens)
{
		/*
		 * We're not capable in ER5 to set a palette. Instead, we find the palette index
		 * which matches the requested color best. 
		 */
	CPalette*	gdiPalette = CPalette::NewDefaultL(iDisplayMode);
	CleanupStack::PushL(gdiPalette);
	for (TUint i = 0; i < totalcolors; i++)
	{
		TRgb	color(palette[3 * i], palette[3 * i + 1], palette[3 * i + 2]);
		
		pens[i] = gdiPalette->NearestIndex(color);

			/*
			 * For low color displays make sure that only black is mapped to black
			 */
		if (CGameBitmap::GetColorDepth(iDisplayMode) == 2)
		{
			if ((color != KRgbBlack) && ((pens[i] == 0) || (pens[i] == 1)))
				pens[i] = 2;
		}
		else if (CGameBitmap::GetColorDepth(iDisplayMode) == 4)
		{
			if ((color != KRgbBlack) && (pens[i] < 5))
				pens[i] = 5;
		}
	}
	CleanupStack::PopAndDestroy();
}



TInt CGameScreen::FindEpocRgbColor(int r, int g, int b)
{
		/*
		 * We're not capable in ER5 to set a palette. Instead, we find the palette index
		 * which matches the requested color best. 
		 */
	CPalette*	gdiPalette = CPalette::NewDefaultL(iDisplayMode);
	CleanupStack::PushL(gdiPalette);
	TRgb	color(r, g, b);
		
	int palette = gdiPalette->NearestIndex(color);
	CleanupStack::PopAndDestroy();

	return palette;
}

TUint CGameScreen::GetColorValue(const TRgb& aColor, const TDisplayMode& aDisplayMode)
{
	switch (aDisplayMode)
	{
		case EColor4K:
			return aColor.Color4K();
			break;
		case EColor16M:
			return aColor.Color16M();
			break;
		case EColor64K:
			return aColor.Color64K();
			break;
		case ERgb:
			return aColor.Value();
			break;
		default:
			ASSERT(EFalse);
	}
	return 0;
}

TInt CGameScreen::AllocateGamePen(const TRgb& aColor)
{
	TUint	colorValue = GetColorValue(aColor, iDisplayMode);
	for (TInt i=0 ; i<iGamePaletteMaxIndex; i++)
		if (iGamePalette[i] == colorValue)
			return i;
	ASSERT(iGamePaletteMaxIndex<KSizeGameColorPalette-1);
	iGamePalette[iGamePaletteMaxIndex++] = colorValue;
	return (iGamePaletteMaxIndex-1);
}


void CGameScreen::AllocateGameColors(TUint totalcolors, const TUint8 *palette, TUint16 *pens, TInt modifiable, const TUint8 *debug_palette,TUint16 *debug_pens)
{
	for (TUint i = 0; i < totalcolors; i++)
	{
		TRgb	color(palette[3 * i], palette[3 * i + 1], palette[3 * i + 2]);
		pens[i] = AllocateGamePen(color);
	}
}


__inline TUint8 MAKE_2BPP_BYTE(TUint8 c1, TUint8 c2, TUint8 c3, TUint8 c4)
{
	return (c4 << 6) | (c3 << 4) | (c2 << 2) | (c1);
}

__inline TUint8 MAKE_4BPP_BYTE(TUint8 c1, TUint8 c2)
{
	return (c2 << 4) | (c1);
}


void CGameScreen::DrawScaledScanLine8bppTo8bpp(TUint* aSrcLine, TUint* aDstLine, TInt aWidth, TBool aMerge)
{
	TInt			pixelCount = 0;
	TUint8*	src = (TUint8*) aSrcLine;
	TUint8*	dst = (TUint8*) aDstLine;

	for (TInt w=0 ; w<aWidth ; w++)
	{
		if (pixelCount == iPixelSkipX)
		{
			pixelCount = 0;
		}
		else
		{
			if (aMerge)
			{
				*dst++ |= *src;
			}
			else
			{
				*dst++ = *src;
			}
			pixelCount++;
		}
		src++;
	}
}

void CGameScreen::DrawDoubleScaledScanLine8bppTo8bpp(TUint* aSrcLine, TUint* aDstLine, TInt aWidth, TBool aMerge)
{
	TInt			pixelCount = 0;
	TUint8*	src = (TUint8*) aSrcLine;
	TUint8*	dst = (TUint8*) aDstLine;

	for (TInt w=0 ; w<aWidth ; w++)
	{
		if (pixelCount == iPixelSkipX)
		{
			pixelCount = 0;
			*(dst-1) |= *src;
			*(dst-2) |= *src;
		}
		else
		{
			if (aMerge)
			{
				*dst++ |= *src;
				*dst++ |= *src;
			}
			else
			{
				*dst++ = *src;
				*dst++ = *src;
			}
			pixelCount++;
		}
		src++;
	}
}

void CGameScreen::DrawScaledScanLine8bppTo2bpp(TUint* aSrcLine, TUint* aDstLine, TInt aWidth, TBool aMerge)
{
	TInt			srcPixCount = 0;
	TInt			dstPixCount = 0;
	TInt			pixelCount = 0;
	TUint			dstPixels = 0;
	TUint			srcPixels = *aSrcLine++;
	for (TInt w=0 ; w<aWidth ; w++)
	{
		dstPixels |= (srcPixels & 3) << (dstPixCount << 1);
		srcPixCount++;
		srcPixels >>= 8;
		if (pixelCount != (iPixelSkipX -1))
			dstPixCount++;

		if (++pixelCount == iPixelSkipX)
			pixelCount=0;

		if (srcPixCount == 4)
		{
			srcPixels = *aSrcLine++;
			srcPixCount=0;
		}
		if (dstPixCount == 16)
		{
			if (aMerge)
			{
				*aDstLine |= dstPixels;
				aDstLine++;
			}
			else
				*aDstLine++ = dstPixels;
			dstPixCount=0;
			dstPixels=0;
		}
	}
}

void CGameScreen::DrawScaledScanLine8bppTo4bpp(TUint* aSrcLine, TUint* aDstLine, TInt aWidth, TBool aMerge)
{
	TInt			srcPixCount = 0;
	TInt			dstPixCount = 0;
	TInt			pixelCount = 0;
	TUint			dstPixels = 0;
	TUint			srcPixels = *aSrcLine++;
	for (TInt w=0 ; w<aWidth ; w++)
	{
		dstPixels |= (srcPixels & 0xf) << (dstPixCount << 2);
		srcPixCount++;
		srcPixels >>= 8;
		if (pixelCount != (iPixelSkipX -1))
			dstPixCount++;

		if (++pixelCount == iPixelSkipX)
			pixelCount=0;

		if (srcPixCount == 4)
		{
			srcPixels = *aSrcLine++;
			srcPixCount=0;
		}
		if (dstPixCount == 8)
		{
			if (aMerge)
			{
				*aDstLine |= dstPixels;
				aDstLine++;
			}
			else
				*aDstLine++ = dstPixels;
			dstPixCount=0;
			dstPixels=0;
		}
	}
}


void CGameScreen::DrawScaledScanLine8bppTo12bpp(TUint* aSrcLine, TUint* aDstLine, TInt aWidth, TBool aMerge)
{
	TInt			pixelCount = 0;
	TUint8*		src = (TUint8*) aSrcLine;
	TUint16*	dst = (TUint16*) aDstLine;

	for (TInt w=0 ; w<aWidth ; w++)
	{
		if (pixelCount == iPixelSkipX)
		{
			pixelCount = 0;
		}
		else
		{
			if (aMerge)
			{
				*dst++ |= iGamePalette[*src];
			}
			else
			{
				*dst++ = iGamePalette[*src];
			}
			pixelCount++;
		}
		src++;
	}
}

void CGameScreen::DirectDrawBitmap8bppTo4bpp(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdVisibleBitmap();

	TUint8* frameBufferLine = iBitmapStartAddress;
	TUint8* bitmapLine;

	TInt	height = (osdBitmap->height > iScreenSize.iHeight) ? iScreenSize.iHeight : osdBitmap->height;
	TInt	width = (osdBitmap->width > iScreenSize.iWidth) ? iScreenSize.iWidth : osdBitmap->width;
	for(TInt h=0 ; h<height ; h++)
	{
		TUint8* frameBuf = frameBufferLine;
		bitmapLine = osdBitmap->line[h];
		for (TInt w=0 ; w<width ; w+=2)
		{
			TUint8	c = MAKE_4BPP_BYTE(bitmapLine[0], bitmapLine[1]);
			*frameBuf++ = c;
			bitmapLine+=2;
		}
		frameBufferLine += iBytesPerScanLine;
	}
}



void CGameScreen::DirectDrawBitmap8bppTo2bpp(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdVisibleBitmap();

	TUint8* frameBufferLine = iBitmapStartAddress;
	TUint8* bitmapLine;

	TInt	height = (osdBitmap->height > iScreenSize.iHeight) ? iScreenSize.iHeight : osdBitmap->height;
	TInt	width = (osdBitmap->width > iScreenSize.iWidth) ? iScreenSize.iWidth : osdBitmap->width;
	for(TInt h=0 ; h<height ; h++)
	{
		TUint8* frameBuf = frameBufferLine;
		bitmapLine = osdBitmap->line[h];
		for (TInt w=0 ; w<width ; w+=4)
		{
			TUint8	c = MAKE_2BPP_BYTE(bitmapLine[0], bitmapLine[1], bitmapLine[2], bitmapLine[3]);
			*frameBuf++ = c;
			bitmapLine+=4;
		}
		frameBufferLine += iBytesPerScanLine;
	}
}

void CGameScreen::DirectDrawBitmap8bppTo12bpp(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdVisibleBitmap();

	TUint8* frameBufferLine = iBitmapStartAddress;
	TUint8* bitmapLine;

	TInt	height = (osdBitmap->height > iScreenSize.iHeight) ? iScreenSize.iHeight : osdBitmap->height;
	TInt	width = (osdBitmap->width > iScreenSize.iWidth) ? iScreenSize.iWidth : osdBitmap->width;
	for(TInt h=0 ; h<height ; h++)
	{
		TUint16* frameBuf = (TUint16* ) frameBufferLine;
		bitmapLine = osdBitmap->line[h];
		for (TInt w=0 ; w<width ; w++)
		{
			*frameBuf++ = (TUint16) iGamePalette[*bitmapLine++];
		}
		frameBufferLine += iBytesPerScanLine;
	}
}



void CGameScreen::DirectDrawBitmap8bppTo8bpp(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdVisibleBitmap();

	TUint8* frameBufferLine = iBitmapStartAddress;
	TUint8* bitmapLine;

	TInt	height = (osdBitmap->height > iScreenSize.iHeight) ? iScreenSize.iHeight : osdBitmap->height;
	TInt	width = (osdBitmap->width > iScreenSize.iWidth) ? iScreenSize.iWidth : osdBitmap->width;
	for(TInt h=0 ; h<height ; h++)
	{
		TUint8* frameBuf = frameBufferLine;
		bitmapLine = osdBitmap->line[h];
		Mem::Copy(frameBuf, bitmapLine, width);
		frameBufferLine += iBytesPerScanLine;
	}
}

void CGameScreen::ScaledDrawBitmap(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdVisibleBitmap();

	TUint8* frameBufLine = iBitmapStartAddress;

	TInt	lineCount = 0;
	
	TInt height = (iPixelSkipY>0) ? osdBitmap->height : ((osdBitmap->height > iScreenSize.iHeight) ? iScreenSize.iHeight : osdBitmap->height);
	TInt width = (iPixelSkipX>0) ? osdBitmap->width : ((osdBitmap->width > iScreenSize.iWidth) ? iScreenSize.iWidth : osdBitmap->width);

	TInt bytesPerBitmapLine = (iPixelSkipX>0) ? iWidth * CGameBitmap::GetBpp(iDisplayMode) / 8 : width * CGameBitmap::GetBpp(iDisplayMode) / 8;
	
	for (TInt h=0 ; h<height ; h++)
	{
		if (lineCount == iPixelSkipY)
		{
			(this->*iScaledScanLineDrawer)((TUint *) osdBitmap->line[h], (TUint*)iScanLine, width, ETrue);
			lineCount = 0;
		}
		else
		{
			(this->*iScaledScanLineDrawer)((TUint *) osdBitmap->line[h], (TUint*)iScanLine, width, EFalse);
			lineCount++;
		}
		if (lineCount != iPixelSkipY)
		{
			Mem::Copy(frameBufLine, iScanLine, bytesPerBitmapLine);
			frameBufLine += iBytesPerScanLine;
		}
	}
}

void CGameScreen::DoubleScaledDrawBitmap(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdVisibleBitmap();

	TUint8* frameBufLine = iBitmapStartAddress;

	TInt	lineCount = 0;
	
	TInt height = (iPixelSkipY>0) ? osdBitmap->height : ((osdBitmap->height > iScreenSize.iHeight) ? iScreenSize.iHeight : osdBitmap->height);
	TInt width = (iPixelSkipX>0) ? osdBitmap->width : ((osdBitmap->width > iScreenSize.iWidth) ? iScreenSize.iWidth : osdBitmap->width);

	TInt bytesPerBitmapLine = (iPixelSkipX>0) ? (iWidth * CGameBitmap::GetBpp(iDisplayMode) / 8) * 2 : (width * CGameBitmap::GetBpp(iDisplayMode) / 8) * 2;

	for (TInt h=0 ; h<height ; h++)
	{
		if (lineCount == iPixelSkipY)
		{
			(this->*iScaledScanLineDrawer)((TUint *) osdBitmap->line[h], (TUint*)iScanLine, width, ETrue);
			lineCount = 0;
		}
		else
		{
			(this->*iScaledScanLineDrawer)((TUint *) osdBitmap->line[h], (TUint*)iScanLine, width, EFalse);
			lineCount++;
		}
		if (lineCount != iPixelSkipY)
		{
			Mem::Copy(frameBufLine, iScanLine, iBytesPerBitmapLine);
			frameBufLine += iBytesPerScanLine;

			if ((iTotalGameFrames < 10) || (iGameFrameCount == iVideoFps))
			{
				Mem::Fill(frameBufLine, bytesPerBitmapLine, 0);
			}
			frameBufLine += iBytesPerScanLine;
		}
	}
}



void CGameScreen::DirectDrawDoubleBitmap8bppTo8bpp(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdVisibleBitmap();

	TUint8* frameBufferLine = iBitmapStartAddress;
	TUint8* bitmapLine;

	TInt	height = osdBitmap->height;
	TInt	width = osdBitmap->width;
	for(TInt h=0 ; h<height ; h++)
	{
		TUint8* frameBuf = frameBufferLine;
		bitmapLine = osdBitmap->line[h];
		for (TInt w=0 ; w<width ; w++)
		{
			*frameBuf++ = *bitmapLine;
			*frameBuf++ = *bitmapLine++;
		}
		frameBufferLine += iBytesPerScanLine;

		if ((iTotalGameFrames < 10) || (iGameFrameCount == iVideoFps))
		{
			Mem::Fill(frameBufferLine, width*2, 0);
		}
		frameBufferLine += iBytesPerScanLine;
	}
}

#if 0
void CGameScreen::DirectDrawDoubleBitmap8bppTo8bpp(CGameBitmap* aGameBitmap)
{
	const struct osd_bitmap*	osdBitmap = aGameBitmap->OsdBitmap();

	TUint8* frameBufferLine = iBitmapStartAddress;
	TUint8* bitmapLine;

	TInt	height = osdBitmap->height;
	TInt	width = osdBitmap->width;
	for(TInt h=0 ; h<height ; h++)
	{
		for (TInt i=0 ; i<2 ; i++)
		{
			TUint8* frameBuf = frameBufferLine;
			bitmapLine = osdBitmap->line[h];
			for (TInt w=0 ; w<width ; w++)
			{
				*frameBuf++ = *bitmapLine;
				*frameBuf++ = *bitmapLine++;
			}
			frameBufferLine += iBytesPerScanLine;
		}
	}
}
#endif

#ifdef __WINS__
CWinsGameScreen::~CWinsGameScreen()
{
	delete iEmulatedScreenBitmap;
}

CWinsGameScreen::CWinsGameScreen(const TSize& aScreenSize, TDisplayMode aDisplayMode)
{
	iScreenSize		= aScreenSize;
	iDisplayMode	= aDisplayMode;
	iHasFrameBuffer = ETrue;
}

void CWinsGameScreen::ConstructL(TInt aWidth, TInt aHeight, TInt aDepth, TOrientation aOrientation, TInt aFps)
{
	iEmulatedScreenBitmap = CGameBitmap::NewL(iScreenSize.iWidth, iScreenSize.iHeight, iDisplayMode);
	iFrameBuffer = (TUint8*) iEmulatedScreenBitmap->OsdVisibleBitmap()->line[0];
	iBytesPerScanLine = (iScreenSize.iWidth * CGameBitmap::GetBpp(iDisplayMode)) / 8;

	CGameScreen::ConstructL(aWidth, aHeight, aDepth, aOrientation, aFps);
};


#endif


/***********************************************
 *
 * GAME WINDOW
 *
 ***********************************************/

CGameWindow::CGameWindow()
{
}

CGameWindow::~CGameWindow()
{
	delete iWsScreen;
	delete iGameBitmapArray;	// we don't know the CGameBitmap objects
}

CGameWindow* CGameWindow::NewL()
{
	CGameWindow*	self = new(ELeave) CGameWindow;
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}


void CGameWindow::ConstructL()
{
	TInt	error;

	iGameBitmapArray = new(ELeave) CGameBitmapArray(KGameBitmapArrayGranularity);

	error = iWsSession.Connect();
	User::LeaveIfError(error);
	iWsScreen=new(ELeave) CWsScreenDevice(iWsSession);
	User::LeaveIfError(iWsScreen->Construct());
	User::LeaveIfError(iWsScreen->CreateContext(iWindowGc));

	iWsWindowGroup=RWindowGroup(iWsSession);
	User::LeaveIfError(iWsWindowGroup.Construct((TUint32)this));
	iWsWindowGroup.SetOrdinalPosition(0);

	iWsWindow=RWindow(iWsSession);
	User::LeaveIfError(iWsWindow.Construct(iWsWindowGroup, (TUint32)this));
	iWsWindow.Activate();
	iWsWindow.SetSize(iWsScreen->SizeInPixels());
	iWsWindow.SetVisible(ETrue);

	StartWServEvents();
}

void CGameWindow::Destruct()
{
	if (iWsWindow.WsHandle())
		iWsWindow.Close();

	if (iWsWindowGroup.WsHandle())
		iWsWindowGroup.Close();

	delete iWindowGc;
	iWindowGc = NULL;

	delete iWsScreen;
	iWsScreen = NULL;

	if (iWsSession.WsHandle())
		iWsSession.Close();
}


void CGameWindow::CreateGameScreenL(TInt aWidth, TInt aHeight, TInt aDepth, TOrientation aOrientation, TInt aFps)
{
	TScreenInfoV01			screenInfo;
	TPckg<TScreenInfoV01>	sI(screenInfo);
	UserSvr::ScreenInfo(sI);

		/* 
		 * Use the scripts in epoc\group\wins to set the EPOC emulator 
		 * to s5mx or s7 appearance
		 */
#ifdef __WINS__
	iGameScreen = new CWinsGameScreen(iWsScreen->SizeInPixels(), iWsScreen->DisplayMode());
#else
	iGameScreen = new CGameScreen(screenInfo, iWsScreen->DisplayMode());
#endif

	iGameScreen->SetWantScaling(ETrue);

	iGameScreen->ConstructL(aWidth, aHeight, aDepth, aOrientation, aFps);
}

void CGameWindow::CreateDisplayL(TInt aWidth, TInt aHeight, TInt aDepth, TOrientation aOrientation, TInt aFps)
{
	delete iGameScreen;
	iGameScreen = NULL;

	CreateGameScreenL(aWidth, aHeight, aDepth, aOrientation, aFps);
	SetupOnScreenControls();
	DrawOnScreenControlsNow();
}


CGameBitmap* CGameWindow::AllocateBitmapL(TInt aWidth, TInt aHeight, TInt aDepth)
{
	CGameBitmap*	gameBitmap;
	
	gameBitmap = CGameBitmap::NewL(aWidth, aHeight, CGameBitmap::GetDisplayMode(aDepth));
	iGameBitmapArray->Append(gameBitmap);

	return gameBitmap;
}



void CGameWindow::SetVisbibleAreaL(const TRect aRect)
{
	for (TInt i=0 ; i<iGameBitmapArray->Count() ; i++)
	{
		(*iGameBitmapArray)[i]->SetVisibleAreaL(aRect);
	}
}

void CGameWindow::MarkInvalid(TInt aXmin, TInt aYmin, TInt aXmax, TInt aYmax)
{
	TSize	size(iWsWindow.Size());
	TRect	rect(aXmin, aYmin, (aXmax == -1) ? size.iWidth : aXmax, (aYmax == -1) ? size.iHeight : aYmax);
}


TInt CGameWindow::SkipThisFrame()
{
	return iGameScreen->SkipThisFrame();
}

void CGameWindow::DrawBitmap(CGameBitmap* aGameBitmap)
{
	if (!iIsPaused)
	{
//		if (input_ui_pressed(IPT_UI_SHOW_FPS))
		{
//			iShowFps = iShowFps ? (TBool) EFalse : (TBool) ETrue;
		}


//		TInt count = iGameScreen->DoubleFrame(iShowFps, aGameBitmap->OsdVisibleBitmap());
		TInt count = 1;
		while(count--)
		{
			if (iGameScreen->HasFrameBuffer())
#ifdef __WINS__
				EmulateTargetDrawBitmap(aGameBitmap);
#else
				iGameScreen->DrawBitmap(aGameBitmap);
#endif
			else
				WsBlitBitmap(aGameBitmap);
		}
	}
}

#ifdef __WINS__
void CGameWindow::EmulateTargetDrawBitmap(CGameBitmap* aGameBitmap)
{
	iWindowGc->Activate(iWsWindow);

	TRect  rect = TRect(iGameScreen->Position(),  iGameScreen->GameBitmapSize());
//	TRect  rect = TRect(0, 0, 640, 480);
	iWsWindow.Invalidate(rect);
	iWsWindow.BeginRedraw(rect);

	if (aGameBitmap)
		iGameScreen->DrawBitmap(aGameBitmap);
	iWindowGc->BitBlt(iGameScreen->Position(), ((CWinsGameScreen*) iGameScreen)->iEmulatedScreenBitmap->FbsBitmap(), rect);
//	iWindowGc->BitBlt(TPoint(0,0), ((CWinsGameScreen*) iGameScreen)->iEmulatedScreenBitmap->FbsBitmap(), rect);
	DrawOnScreenControls();

	iWsWindow.EndRedraw();
	iWindowGc->Deactivate();
	iWsSession.Flush();
}
#endif


void CGameWindow::WsBlitBitmap(CGameBitmap* aGameBitmap)
{
	iWindowGc->Activate(iWsWindow);

	TRect  rect = TRect(iWsWindow.Size());
	iWsWindow.Invalidate(rect);
	iWsWindow.BeginRedraw(rect);

	if (aGameBitmap)
	{
		iWindowGc->BitBlt(iGameScreen->Position(), aGameBitmap->FbsBitmap());
	}

	iWsWindow.EndRedraw();
	iWindowGc->Deactivate();
	iWsSession.Flush();
}




void CGameWindow::AllocateColor(TUint totalcolors, const TUint8 *palette, TUint16 *pens, TInt modifiable, const TUint8 *debug_palette,TUint16 *debug_pens)
{
	iGameScreen->AllocateColors(totalcolors, palette, pens, modifiable, debug_palette,debug_pens);
}



void CGameWindow::CloseDisplay()
{
	delete iGameScreen;
	iGameScreen = 0;

	CancelWServEvents();

	Destruct();
}

TInt TranslateKeyEvent(TInt aKeyCode)
{
#if 0
	if (options.rol)
	{
		if ((aKeyCode>= EKeyLeftArrow) && (aKeyCode <= EKeyDownArrow))
		{
			const TInt	mappedKeys[] = {EKeyUpArrow, EKeyDownArrow, EKeyRightArrow, EKeyLeftArrow};
			return mappedKeys[aKeyCode - EKeyLeftArrow];
		}
	}
	else if (options.ror)
	{
		if ((aKeyCode>= EKeyLeftArrow) && (aKeyCode <= EKeyDownArrow))
		{
			const TInt	mappedKeys[] = {EKeyDownArrow, EKeyUpArrow, EKeyLeftArrow, EKeyRightArrow};
			return mappedKeys[aKeyCode - EKeyLeftArrow];
		}
	}
#endif
	return aKeyCode;
}

TInt CGameWindow::GetOnScreenKey(const TPoint aPosition)
{
	for (TInt i=0 ; i<MAX_ONSCREENCONTROLS ; i++)
	{
		if (iOnScreenControlList[i].iRect.Contains(aPosition))
			return iOnScreenControlList[i].iKey;
	}
	return -1;
}

void CGameWindow::HandlePointerEventL(const TPointerEvent& aPointerEvent)
{
	if (aPointerEvent.iType==TPointerEvent::EButton1Down)
	{
		iLastOnScreenKey = GetOnScreenKey(aPointerEvent.iPosition);
	}
	else if (aPointerEvent.iType==TPointerEvent::EButton1Up)
	{
		iLastOnScreenKey = -1;
	}
}


void CGameWindow::KeyDownEvent(TInt aScanCode)
{
	if (iKeyboardObserver)
	{
		iKeyboardObserver->KeyDownEvent(aScanCode);
	}
	else if (iKeyPressLevel < MAX_KEYPRESS)
	{
		iKeyPress[iKeyPressLevel].iScanCode = aScanCode;
		iKeyPressLevel++;
	}
}

void CGameWindow::KeyPressEvent(TInt aKeyCode)
{
	iKeyPress[iKeyPressLevel-1].iKeyCode = aKeyCode;
}

void CGameWindow::KeyUpEvent(TInt aScanCode)
{
	if (iKeyboardObserver)
	{
		iKeyboardObserver->KeyUpEvent(aScanCode);
	}
	else
	{
		for (TInt i=iKeyPressLevel-1 ; i>=0 ; i--)
		{
			if (iKeyPress[i].iScanCode == aScanCode)
			{
				iKeyPress[i].iScanCode = iKeyPress[i].iKeyCode = 0;
				break;
			}
		}

		while(iKeyPressLevel>0 && (iKeyPress[iKeyPressLevel-1].iKeyCode == 0))
			iKeyPressLevel--;
	}
}

TInt CGameWindow::CurrentKey()
{
	if (iKeyPressLevel == 0)
		return 0;
	return iKeyPress[iKeyPressLevel-1].iKeyCode;
}


void CGameWindow::HandleWsEvent(const TWsEvent& aWsEvent)
{
//	DPRINT((HERE("CGameWindow::HandleWsEvent, type=%d"), aWsEvent.Type()));

	if (aWsEvent.Type() == EEventKeyDown)
	{
		KeyDownEvent(aWsEvent.Key()->iScanCode);
	} 
	else if (aWsEvent.Type() == EEventKey)
	{
		KeyPressEvent(TranslateKeyEvent(aWsEvent.Key()->iCode));
		if (aWsEvent.Key()->iCode == '+')
		{
			sManualDiffFrames++;
		}
		else if (aWsEvent.Key()->iCode == '-')
		{
			if (sManualDiffFrames > 0)
				sManualDiffFrames--;
		}
	} 
	else if (aWsEvent.Type() == EEventKeyUp)
	{
		KeyUpEvent(aWsEvent.Key()->iScanCode);
	} 
	else if (aWsEvent.Type() == EEventPointer)
	{
		HandlePointerEventL(*aWsEvent.Pointer());
	}
	else if(aWsEvent.Type() == EEventFocusLost)
	{
		if (!iIsPaused)
		{
			iLastOnScreenKey = 'p';		// simulate a pause
		}
	}
	else if(aWsEvent.Type() == EEventFocusGained)
	{
//		if (iIsPaused)
//			WsBlitBitmap((CGameBitmap*)Machine->scrbitmap->_private);		
	}
}


const KSmallAxis= 50;
const KLongAxis	= 100;

void CGameWindow::SetupOnScreenControls()
{
	TInt	heightDiv2 = iGameScreen->ScreenSize().iHeight / 2;
	TInt	width = iGameScreen->ScreenSize().iWidth;

	const TRect	horElipse(0, 0, KLongAxis, KSmallAxis);
	const TRect	verElipse(0, 0, KSmallAxis, KLongAxis);

	switch (iGameScreen->Orientation())
	{
		case EOrientationRotRight:
			iOnScreenControlList[0].iKey	= EKeyLeftArrow;
			iOnScreenControlList[0].iRect	= verElipse;
			iOnScreenControlList[0].iRect.Move(KLongAxis, heightDiv2 - (KSmallAxis / 2) - KLongAxis);

			iOnScreenControlList[1].iKey	= EKeyRightArrow;
			iOnScreenControlList[1].iRect	= verElipse;
			iOnScreenControlList[1].iRect.Move(KLongAxis, heightDiv2 + (KSmallAxis / 2));

			iOnScreenControlList[2].iKey	= EKeyUpArrow;
			iOnScreenControlList[2].iRect	= horElipse;
			iOnScreenControlList[2].iRect.Move(KLongAxis + KSmallAxis, heightDiv2 - (KSmallAxis / 2));

			iOnScreenControlList[3].iKey	= EKeyDownArrow;
			iOnScreenControlList[3].iRect	= horElipse;
			iOnScreenControlList[3].iRect.Move(0, heightDiv2 - (KSmallAxis / 2));
			break;

		case EOrientationRotLeft:
			iOnScreenControlList[0].iKey	= EKeyRightArrow;
			iOnScreenControlList[0].iRect	= verElipse;
			iOnScreenControlList[0].iRect.Move(width - KSmallAxis - KLongAxis, heightDiv2 - (KSmallAxis / 2) - KLongAxis);

			iOnScreenControlList[1].iKey	= EKeyLeftArrow;
			iOnScreenControlList[1].iRect	= verElipse;
			iOnScreenControlList[1].iRect.Move(width - KSmallAxis - KLongAxis, heightDiv2 + (KSmallAxis / 2));

			iOnScreenControlList[2].iKey	= EKeyUpArrow;
			iOnScreenControlList[2].iRect	= horElipse;
			iOnScreenControlList[2].iRect.Move(width - KLongAxis - (KLongAxis + KSmallAxis), heightDiv2 - (KSmallAxis / 2));

			iOnScreenControlList[3].iKey	= EKeyDownArrow;
			iOnScreenControlList[3].iRect	= horElipse;
			iOnScreenControlList[3].iRect.Move(width - KLongAxis, heightDiv2 - (KSmallAxis / 2));
			break;
	}
}

void CGameWindow::DrawOnScreenControlsNow()
{
	iWindowGc->Activate(iWsWindow);

	TRect  rect = TRect(iWsWindow.Size());
	iWsWindow.Invalidate(rect);
	iWsWindow.BeginRedraw(rect);

	DrawOnScreenControls();

	iWsWindow.EndRedraw();
	iWindowGc->Deactivate();
	iWsSession.Flush();
}

void CGameWindow::DrawOnScreenControls()
{
	iWindowGc->SetBrushStyle(CGraphicsContext::ESolidBrush);
	iWindowGc->SetBrushColor(KRgbBlack);
	for (TInt i=0 ; i<MAX_ONSCREENCONTROLS ; i++)
		if (iOnScreenControlList->iKey) iWindowGc->DrawEllipse(iOnScreenControlList[i].iRect);
}

void CGameWindow::HandleRedrawEvent(const TWsRedrawEvent& iRedrawEvent)
{
//	DPRINT((HERE("CGameWindow::HandleRedrawEvent, handle=%d"), iRedrawEvent.Handle()));
	
	// Server initiated drawing not needed here
//	WsBlitBitmap((CGameBitmap*)Machine->scrbitmap->_private);
}


void CGameWindow::StartWServEvents()
{
	iWsEventStatus = KRequestPending;
	iWsSession.EventReady(&iWsEventStatus);

	iRedrawEventStatus = KRequestPending;
	iWsSession.RedrawReady(&iRedrawEventStatus);
}

void CGameWindow::CancelWServEvents()
{
	if (iWsEventStatus != KRequestPending)
		iWsSession.EventReadyCancel();

	if (iRedrawEventStatus != KRequestPending)
		iWsSession.RedrawReadyCancel();
}


void CGameWindow::PollWServEvents()
{
	if (iWsEventStatus != KRequestPending)
	{
		iWsSession.GetEvent(iWsEvent);
		HandleWsEvent(iWsEvent);
		iWsEventStatus = KRequestPending;
		iWsSession.EventReady(&iWsEventStatus);
	}
}

TBool CGameWindow::IsKeyPressed(TInt aKey)
{
	PollWServEvents();

	TBool	isPressed = (aKey == CurrentKey()) || (aKey == iLastOnScreenKey);

	return isPressed;
}

void CGameWindow::Pause(TBool aPause)
{
	iIsPaused = aPause;
}

void CGameWindow::SetKeyboardObserver(MKeyboardObserver* aKeyboardObserver)
{
	iKeyboardObserver = aKeyboardObserver;
}
